using System.Collections;
using System.Collections.Generic;
using System;
using UnityEngine;

/*
 * v2
 * 
 * - fix toggles not resetting physics
 * - fix left and right shoulder switched
 * 
 */

namespace Eosin
{
    public class IronGrip : MVRScript
    {
        JSONStorableFloat uiStrengthFactor;
        public float strengthFactor { get { return uiStrengthFactor.val; } set { uiStrengthFactor.val = value; } }

        JSONStorableFloat uiLimbStrengthFactor;
        public float limbStrengthFactor { get { return uiLimbStrengthFactor.val; } set { uiLimbStrengthFactor.val = value; } }

        JSONStorableFloat fingersCloserTogether;
        JSONStorableFloat fingerFlexOffset;
        JSONStorableFloat fingerColliderScale;
        JSONStorableFloat fingerColliderOffset;
        JSONStorableFloat fingerMassScaleFactor;
        JSONStorableBool visualizeColliders;
        JSONStorableBool overrideFingerPhysics;
        JSONStorableBool useFingerDefaultsAsMin;
        JSONStorableBool overrideHandPhysics;
        JSONStorableBool overrideElbowPhysics;
        JSONStorableBool overrideShoulderPhysics;
        JSONStorableBool leftSide;
        JSONStorableBool rightSide;

        JSONStorableFloat minFingerSpring;
        JSONStorableFloat maxFingerSpring;
        JSONStorableFloat minFingerDamper;
        JSONStorableFloat maxFingerDamper;
        JSONStorableFloat minFingerLimitSpring;
        JSONStorableFloat maxFingerLimitSpring;
        JSONStorableFloat minFingerLimitDamper;
        JSONStorableFloat maxFingerLimitDamper;
        JSONStorableFloat minFingerMaxForce;
        JSONStorableFloat maxFingerMaxForce;

        FreeControllerV3 lHandControl;
        FreeControllerV3 rHandControl;
        FreeControllerV3 lElbowControl;
        FreeControllerV3 rElbowControl;
        FreeControllerV3 lShoulderControl;
        FreeControllerV3 rShoulderControl;

        FreeControllerV3[] controllers;

        JSONStorableFloat minHoldPositionSpring;
        JSONStorableFloat maxHoldPositionSpring;
        JSONStorableFloat minHoldPositionDamper;
        JSONStorableFloat maxHoldPositionDamper;

        JSONStorableFloat minHoldRotationSpring;
        JSONStorableFloat maxHoldRotationSpring;
        JSONStorableFloat minHoldRotationDamper;
        JSONStorableFloat maxHoldRotationDamper;

        JSONStorableFloat minJointDriveSpring;
        JSONStorableFloat maxJointDriveSpring;
        JSONStorableFloat minJointDriveDamper;
        JSONStorableFloat maxJointDriveDamper;
        JSONStorableFloat minMaxForce;
        JSONStorableFloat maxMaxForce;

        struct Range
        {
            public float min
            {
                get { return _min; }
                set
                {
                    if (value >= _max)
                    {
                        _min = _max;
                    }

                    else _min = value;
                }
            }
            public float max
            {
                get { return _max; }
                set
                {
                    if (value <= _min)
                    {
                        _max = _min;
                    }
                    else _max = value;
                }
            }
            float _min, _max;
            public float at(float v)
            {
                return _min + v * (_max - _min);
            }
        }

        Range holdPositionSpringRange = new Range();
        Range holdPositionDamperRange = new Range();
        Range holdRotationSpringRange = new Range();
        Range holdRotationDamperRange = new Range();
        Range jointDriveSpringRange = new Range();
        Range jointDriveDamperRange = new Range();
        Range jointDriveMaxForceRange = new Range();

        Range fingerSpringRange = new Range();
        Range fingerDamperRange = new Range();
        Range fingerLimitSpringRange = new Range();
        Range fingerLimitDamperRange = new Range();
        Range fingerMaxForceRange = new Range();

        void SetRanges()
        {
            holdPositionSpringRange.min = minHoldPositionSpring.val;
            holdPositionSpringRange.max = maxHoldPositionSpring.val;
            holdPositionDamperRange.min = minHoldPositionDamper.val;
            holdPositionDamperRange.max = maxHoldPositionDamper.val;
            holdRotationSpringRange.min = minHoldRotationSpring.val;
            holdRotationSpringRange.max = maxHoldRotationSpring.val;
            holdRotationDamperRange.min = minHoldRotationDamper.val;
            holdRotationDamperRange.max = maxHoldRotationDamper.val;
            jointDriveSpringRange.min = minJointDriveSpring.val;
            jointDriveSpringRange.max = maxJointDriveSpring.val;
            jointDriveDamperRange.min = minJointDriveDamper.val;
            jointDriveDamperRange.max = maxJointDriveDamper.val;
            jointDriveMaxForceRange.min = minMaxForce.val;
            jointDriveMaxForceRange.max = maxMaxForce.val;
            fingerSpringRange.min = minFingerSpring.val;
            fingerSpringRange.max = maxFingerSpring.val;
            fingerDamperRange.min = minFingerDamper.val;
            fingerDamperRange.max = maxFingerDamper.val;
            fingerLimitSpringRange.min = minFingerLimitSpring.val;
            fingerLimitSpringRange.max = maxFingerLimitSpring.val;
            fingerLimitDamperRange.min = minFingerLimitDamper.val;
            fingerLimitDamperRange.max = maxFingerLimitDamper.val;
            fingerMaxForceRange.min = minFingerMaxForce.val;
            fingerMaxForceRange.max = maxFingerMaxForce.val;
        }

        void GetControllers()
        {
            lHandControl = (FreeControllerV3)containingAtom.GetStorableByID("lHandControl");
            rHandControl = (FreeControllerV3)containingAtom.GetStorableByID("rHandControl");
            lElbowControl = (FreeControllerV3)containingAtom.GetStorableByID("lElbowControl");
            rElbowControl = (FreeControllerV3)containingAtom.GetStorableByID("rElbowControl");
            lShoulderControl = (FreeControllerV3)containingAtom.GetStorableByID("lArmControl");
            rShoulderControl = (FreeControllerV3)containingAtom.GetStorableByID("rArmControl");
            controllers = new FreeControllerV3[] { lHandControl, rHandControl, lElbowControl, rElbowControl, lShoulderControl, rShoulderControl };
        }

        public override void Init()
        {
            GetControllers();
            RegisterUIElements();
            CreateUI();
            SetRanges();
            GetBones();
            ApplyFingerToePhysics();
            ApplyFingerColliderSettings();
            ApplyPoseMorph();
            ApplyLimbPhysics();
        }

        void OnDestroy()
        {
            foreach (var pair in colliderMarkers)
                Destroy(pair.Value);

            ResetPoseMorph(true);
            ResetFingerToePhysics(true);
            ResetLimbPhysics(true, true);
            ResetFingerColliderSettings(true);
        }

        HashSet<DAZBone> fingerToeBones;
        string[] handFootBoneNames = new string[] { "Thumb", "Index", "Mid", "Ring", "Pinky", "Carpal", "SmallToe", "BigToe" };

        public void GetBone(Transform bone)
        {
            bool bHandFootBone = false;
            foreach (string boneName in handFootBoneNames)
            {
                if (bone.name.Contains(boneName))
                {
                    bHandFootBone = true;
                    break;
                }
            }
            if (bHandFootBone)
            {
                fingerToeBones.Add(bone.gameObject.GetComponent<DAZBone>());
            }

            for (var i = 0; i < bone.childCount; i++)
            {
                var child = bone.GetChild(i);
                if (child.gameObject.GetComponent<DAZBone>() != null)
                {
                    GetBone(child);
                }
            }
        }

        void GetBones()
        {
            foreach (var parent in containingAtom.gameObject.GetComponentsInChildren<DAZBones>())
            {
                if (parent.name == "Genesis2Female" || parent.name == "Genesis2Male")
                {
                    fingerToeBones = new HashSet<DAZBone>();
                    GetBone(parent.transform);
                }
            }
        }

        Dictionary<DAZBone, JointPhysics> defaultJointPhysics;

        struct JointPhysics
        {
            public SoftJointLimitSpring angularXLimitSpring;
            public SoftJointLimitSpring angularYZLimitSpring;
            public JointDrive angularXDrive;
            public JointDrive angularYZDrive;
            public JointDrive slerpDrive;
            public float mass;
            public float massScale;
            public float connectedMassScale;
        }

        Dictionary<FreeControllerV3, ControllerPhysics> defaultControllerPhysics;

        struct ControllerPhysics
        {
            public float holdPositionSpring;
            public float holdPositionDamper;
            public float holdRotationSpring;
            public float holdRotationDamper;
            public float complyPositionSpring;
            public float complyPositionDamper;
            public float complyRotationSpring;
            public float complyRotationDamper;
            public float jointDriveSpring;
            public float jointDriveDamper;
            public float jointDriveMaxForce;
        }

        void ResetLimbPhysics(bool resetBothSides = false, bool resetAllJoints = false)
        {
            foreach (FreeControllerV3 controller in controllers)
            {
                if(!resetBothSides)
                {
                    bool leftSideBone = (controller == lHandControl || controller == lElbowControl || controller == lShoulderControl);
                    if ((leftSideBone && leftSide.val) || (!leftSideBone && rightSide.val))
                        continue;
                }

                if(!resetAllJoints)
                {
                    if (((controller == lHandControl || controller == rHandControl) && overrideHandPhysics.val) ||
                    ((controller == lElbowControl || controller == rElbowControl) && overrideElbowPhysics.val) ||
                    ((controller == lShoulderControl || controller == rShoulderControl) && overrideShoulderPhysics.val))
                        continue;
                }

                ControllerPhysics phys = defaultControllerPhysics[controller];
                controller.RBHoldPositionSpring = phys.holdPositionSpring;
                controller.RBHoldPositionDamper = phys.holdPositionDamper;
                controller.RBHoldRotationSpring = phys.holdRotationSpring;
                controller.RBHoldRotationSpring = phys.holdRotationSpring;
                controller.jointRotationDriveSpring = phys.jointDriveSpring;
                controller.jointRotationDriveDamper = phys.jointDriveDamper;
                controller.jointRotationDriveMaxForce = phys.jointDriveMaxForce;
                controller.RBComplyPositionSpring = phys.holdPositionSpring;
                controller.RBComplyPositionDamper = phys.holdPositionDamper;
                controller.RBComplyRotationSpring = phys.holdRotationSpring;
                controller.RBComplyRotationSpring = phys.holdRotationSpring;
                controller.RBComplyJointRotationDriveSpring = phys.jointDriveSpring;
            }
        }

        void ApplyLimbPhysics()
        {
            bool needDefaults = false;

            if (defaultControllerPhysics == null)
            {
                defaultControllerPhysics = new Dictionary<FreeControllerV3, ControllerPhysics>();
                needDefaults = true;
            }

            if (!needDefaults && !overrideElbowPhysics.val && !overrideHandPhysics.val && !overrideShoulderPhysics.val)
                return;

            foreach (FreeControllerV3 controller in controllers)
            {
                if (needDefaults)
                {
                    ControllerPhysics phys = new ControllerPhysics();
                    phys.holdPositionSpring = controller.RBHoldPositionSpring;
                    phys.holdPositionDamper = controller.RBHoldPositionDamper;
                    phys.holdRotationSpring = controller.RBHoldRotationSpring;
                    phys.holdRotationDamper = controller.RBHoldRotationDamper;
                    phys.jointDriveSpring = controller.jointRotationDriveSpring;
                    phys.jointDriveDamper = controller.jointRotationDriveDamper;
                    phys.jointDriveMaxForce = controller.jointRotationDriveMaxForce;
                    defaultControllerPhysics[controller] = phys;
                }
                bool leftSideBone = (controller == lHandControl || controller == lElbowControl || controller == lShoulderControl);
                if ((leftSideBone && !leftSide.val) || (!leftSideBone && !rightSide.val))
                    continue;

                if (((controller == lHandControl || controller == rHandControl) && overrideHandPhysics.val) ||
                        ((controller == lElbowControl || controller == rElbowControl) && overrideElbowPhysics.val) ||
                        ((controller == lShoulderControl || controller == rShoulderControl) && overrideShoulderPhysics.val))
                {
                    float factor = uiStrengthFactor.val / 100f;
                    factor *= uiLimbStrengthFactor.val / 100f;
                    float holdPositionSpring = holdPositionSpringRange.at(factor);
                    float holdPositionDamper = holdPositionDamperRange.at(factor);
                    float holdRotationSpring = holdRotationSpringRange.at(factor);
                    float holdRotationDamper = holdRotationDamperRange.at(factor);
                    float jointDriveSpring = jointDriveSpringRange.at(factor);
                    float jointDriveDamper = jointDriveDamperRange.at(factor);
                    float jointDriveMaxForce = jointDriveMaxForceRange.at(factor);

                    controller.RBHoldPositionSpring = holdPositionSpring;
                    controller.RBHoldPositionDamper = holdPositionDamper;
                    controller.RBHoldRotationSpring = holdRotationSpring;
                    controller.RBHoldRotationSpring = holdRotationSpring;
                    controller.jointRotationDriveSpring = jointDriveSpring;
                    controller.jointRotationDriveDamper = jointDriveDamper;
                    controller.jointRotationDriveMaxForce = jointDriveMaxForce;

                    controller.RBComplyPositionSpring = holdPositionSpring;
                    controller.RBComplyPositionDamper = holdPositionDamper;
                    controller.RBComplyRotationSpring = holdRotationSpring;
                    controller.RBComplyRotationSpring = holdRotationSpring;
                    controller.RBComplyJointRotationDriveSpring = jointDriveSpring;
                }
            }
        }

        void ApplyFingerToePhysics()
        {
            try
            {
                bool needDefaults = false;
                if (defaultJointPhysics == null)
                {
                    defaultJointPhysics = new Dictionary<DAZBone, JointPhysics>();
                    needDefaults = true;
                }

                if (!overrideFingerPhysics.val && !needDefaults)
                    return;

                float factor = uiStrengthFactor.val / 100f;
                float angularSpring = fingerSpringRange.at(factor);
                float angularDamper = fingerDamperRange.at(factor);
                float angularLimitSpring = fingerLimitSpringRange.at(factor);
                float angularLimitDamper = fingerLimitDamperRange.at(factor);
                float maximumForce = fingerMaxForceRange.at(factor);

                foreach (var bone in fingerToeBones)
                {
                    ConfigurableJoint joint = bone.gameObject.GetComponent<ConfigurableJoint>();
                    Rigidbody rb = bone.gameObject.GetComponent<Rigidbody>();

                    if (needDefaults)
                    {
                        /*    
                        string[] values = new string[]
                        {
                            joint.angularXDrive.positionSpring.ToString(), joint.angularXDrive.positionDamper.ToString(),
                            joint.angularYZDrive.positionSpring.ToString(), joint.angularYZDrive.positionDamper.ToString(),
                            joint.slerpDrive.positionSpring.ToString(), joint.slerpDrive.positionDamper.ToString(),
                            joint.angularXLimitSpring.spring.ToString(), joint.angularXLimitSpring.damper.ToString(),
                            joint.angularYZLimitSpring.spring.ToString(), joint.angularYZLimitSpring.damper.ToString(),
                            joint.angularXDrive.maximumForce.ToString(), joint.angularYZDrive.maximumForce.ToString()
                        };
                        SuperController.LogMessage(joint.name + ": " + string.Join(", ", values));*/

                        // index
                        // 2, 0.1,
                        // 8, 0.1,
                        // 2, 0.1,
                        // 0, 0, 0, 0, limits
                        // 5, 5 max force

                        JointPhysics phys = new JointPhysics();
                        phys.angularXLimitSpring = joint.angularXLimitSpring;
                        phys.angularYZLimitSpring = joint.angularYZLimitSpring;
                        phys.angularXDrive = joint.angularXDrive;
                        phys.angularYZDrive = joint.angularYZDrive;
                        phys.slerpDrive = joint.slerpDrive;
                        phys.massScale = joint.massScale;
                        phys.connectedMassScale = joint.connectedMassScale;
                        phys.mass = rb.mass;
                        defaultJointPhysics[bone] = phys;
                    }

                    if (!overrideFingerPhysics.val)
                        continue;

                    bool leftSideBone = bone.name.StartsWith("l");
                    if ((leftSideBone && !leftSide.val) || (!leftSideBone && !rightSide.val))
                        continue;

                    float minScale = defaultJointPhysics[bone].massScale;
                    float maxScale = Mathf.Max(minScale, fingerMassScaleFactor.val / 100f);
                    float appliedScale = Mathf.Lerp(minScale, maxScale, factor);

                    // joint.massScale = defaultBonePhysics[bone].massScale * appliedScale;
                    // joint.connectedMassScale = defaultBonePhysics[bone].connectedMassScale * appliedScale;
                    rb.mass = defaultJointPhysics[bone].mass * appliedScale;

                    if (!useFingerDefaultsAsMin.val)
                    {
                        JointDrive myJointDrive = new JointDrive();
                        myJointDrive.positionDamper = angularDamper;
                        myJointDrive.positionSpring = angularSpring;
                        myJointDrive.maximumForce = maximumForce;

                        SoftJointLimitSpring myLimitSpring = new SoftJointLimitSpring();
                        myLimitSpring.damper = angularLimitDamper;
                        myLimitSpring.spring = angularLimitSpring;

                        joint.xDrive = myJointDrive;
                        joint.yDrive = myJointDrive;
                        joint.zDrive = myJointDrive;

                        joint.slerpDrive = myJointDrive; // toes, carpals and Thumb1
                        joint.angularXDrive = myJointDrive; // fingers
                        //if (myJointDrive.positionSpring >= defaultBonePhysics[bone].angularYZDrive.positionSpring && myJointDrive.positionDamper >= defaultBonePhysics[bone].angularYZDrive.positionDamper) // attempt prevent finger/toe rotations around the primary bone axis at low power
                        joint.angularYZDrive = myJointDrive;

                        joint.angularXLimitSpring = myLimitSpring;
                        //if (myLimitSpring.spring >= defaultBonePhysics[bone].angularYZLimitSpring.spring && myLimitSpring.damper >= defaultBonePhysics[bone].angularYZLimitSpring.damper) // attempt prevent finger/toe rotations around the primary bone axis at low power
                        joint.angularYZLimitSpring = myLimitSpring;
                    }
                    else
                    {
                        JointPhysics def = defaultJointPhysics[bone];

                        JointDrive slerpDrive = new JointDrive();
                        slerpDrive.positionSpring = Mathf.Lerp(def.slerpDrive.positionSpring, fingerSpringRange.max, factor);
                        slerpDrive.positionDamper = Mathf.Lerp(def.slerpDrive.positionDamper, fingerDamperRange.max, factor);
                        slerpDrive.maximumForce = Mathf.Lerp(def.slerpDrive.maximumForce, fingerMaxForceRange.max, factor);
                        JointDrive angularXDrive = new JointDrive();
                        angularXDrive.positionSpring = Mathf.Lerp(def.angularXDrive.positionSpring, fingerSpringRange.max, factor);
                        angularXDrive.positionDamper = Mathf.Lerp(def.angularXDrive.positionDamper, fingerDamperRange.max, factor);
                        angularXDrive.maximumForce = Mathf.Lerp(def.angularXDrive.maximumForce, fingerMaxForceRange.max, factor);
                        JointDrive angularYZDrive = new JointDrive();
                        angularYZDrive.positionSpring = Mathf.Lerp(def.angularYZDrive.positionSpring, fingerSpringRange.max, factor);
                        angularYZDrive.positionDamper = Mathf.Lerp(def.angularYZDrive.positionDamper, fingerDamperRange.max, factor);
                        angularYZDrive.maximumForce = Mathf.Lerp(def.angularYZDrive.maximumForce, fingerMaxForceRange.max, factor);
                        SoftJointLimitSpring angularXLimitSpring = new SoftJointLimitSpring();
                        angularXLimitSpring.damper = Mathf.Lerp(def.angularXLimitSpring.damper, fingerLimitDamperRange.max, factor);
                        angularXLimitSpring.spring = Mathf.Lerp(def.angularXLimitSpring.spring, fingerLimitSpringRange.max, factor);
                        SoftJointLimitSpring angularYZLimitSpring = new SoftJointLimitSpring();
                        angularYZLimitSpring.damper = Mathf.Lerp(def.angularYZLimitSpring.damper, fingerLimitDamperRange.max, factor);
                        angularYZLimitSpring.spring = Mathf.Lerp(def.angularYZLimitSpring.spring, fingerLimitSpringRange.max, factor);

                        joint.slerpDrive = slerpDrive;
                        joint.angularXDrive = angularXDrive;
                        joint.angularYZDrive = angularYZDrive;
                        joint.angularXLimitSpring = angularXLimitSpring;
                        joint.angularYZLimitSpring = angularYZLimitSpring;
                    }
                }
            }
            catch (Exception e)
            {
                SuperController.LogError(e.ToString());
            }

        }

        void ResetFingerToePhysics(bool resetBothSides = false)
        {
            if (defaultJointPhysics == null || fingerToeBones == null)
                return;
            foreach (var bone in fingerToeBones)
            {
                if (!resetBothSides)
                {
                    bool leftSideBone = bone.name.StartsWith("l");
                    if ((leftSideBone && leftSide.val) || (!leftSideBone && rightSide.val))
                        continue;
                }

                var joint = bone.gameObject.GetComponent<ConfigurableJoint>();
                if (joint == null || !defaultJointPhysics.ContainsKey(bone))
                    continue;

                var defaults = defaultJointPhysics[bone];
                joint.angularXLimitSpring = defaults.angularXLimitSpring;
                joint.angularYZLimitSpring = defaults.angularYZLimitSpring;
                joint.angularXDrive = defaults.angularXDrive;
                joint.angularYZDrive = defaults.angularYZDrive;
                joint.slerpDrive = defaults.slerpDrive;
                joint.massScale = defaults.massScale;
                joint.connectedMassScale = defaults.connectedMassScale;
                joint.gameObject.GetComponent<Rigidbody>().mass = defaults.mass;
            }
        }

        void ResetPoseMorph(bool resetBothSides = false)
        {
            if (fingerToeBones == null)
                return;
            foreach (var bone in fingerToeBones)
            {
                if (!resetBothSides)
                {
                    bool leftSideBone = bone.name.StartsWith("l");
                    if ((leftSideBone && leftSide.val) || (!leftSideBone && rightSide.val))
                        continue;
                }

                bone.SetBoneXRotation("IronGrip", 0f);
                bone.SetBoneYRotation("IronGrip", 0f);
                bone.SetBoneZRotation("IronGrip", 0f);
                bone.SyncMorphBoneRotations(false);
            }
        }

        Dictionary<CapsuleCollider, float> defaultRadii;
        Dictionary<CapsuleCollider, Vector3> defaultLocalPositions = new Dictionary<CapsuleCollider, Vector3>();

        void ResetFingerColliderSettings(bool resetBothSides = false)
        {
            if (fingerToeBones == null || defaultRadii == null)
                return;
            foreach (var bone in fingerToeBones)
            {
                if (bone.name.Contains("Toe"))
                    continue;

                if (!resetBothSides)
                {
                    bool leftSideBone = bone.name.StartsWith("l");
                    if ((leftSideBone && leftSide.val) || (!leftSideBone && rightSide.val))
                        continue;
                }

                foreach (Transform child in bone.transform)
                {
                    foreach (CapsuleCollider collider in child.gameObject.GetComponents<CapsuleCollider>())
                    {
                        collider.radius = defaultRadii[collider];
                        collider.transform.localPosition = defaultLocalPositions[collider];
                        float radius = 0.85f * collider.radius * 2f;
                        colliderMarkers[collider].transform.localScale = (fingerColliderScale.val / 100f) * new Vector3(radius, 0.85f * collider.height / 2, radius);
                    }
                }
            }
        }

        Dictionary<CapsuleCollider, GameObject> colliderMarkers = new Dictionary<CapsuleCollider, GameObject>();

        void ApplyFingerColliderSettings()
        {
            if (fingerToeBones == null)
                return;

            bool needDefaults = false;
            if (defaultRadii == null)
            {
                defaultRadii = new Dictionary<CapsuleCollider, float>();
                needDefaults = true;
            }
            foreach (var bone in fingerToeBones)
            {
                if (bone.name.Contains("Toe"))
                    continue;

                bool leftSideBone = bone.name.StartsWith("l");
                if ((leftSideBone && !leftSide.val) || (!leftSideBone && !rightSide.val))
                    continue;

                foreach (Transform child in bone.transform)
                {
                    foreach (CapsuleCollider collider in child.gameObject.GetComponents<CapsuleCollider>())
                    {
                        if (needDefaults)
                        {
                            defaultRadii[collider] = collider.radius;
                            defaultLocalPositions[collider] = collider.transform.localPosition;
                        }
                        collider.radius = defaultRadii[collider] * fingerColliderScale.val / 100f;
                        float offset = fingerColliderOffset.val / 100f;
                        if (bone.name.Contains("Thumb"))
                            offset = -offset;
                        collider.transform.localPosition = defaultLocalPositions[collider] + Vector3.up * offset;

                        float radius = 0.85f * collider.radius * 2f;
                        if (!colliderMarkers.ContainsKey(collider))
                        {
                            GameObject n = GameObject.CreatePrimitive(PrimitiveType.Capsule);
                            DestroyImmediate(n.GetComponent<Rigidbody>());
                            DestroyImmediate(n.GetComponent<CapsuleCollider>());

                            n.transform.localScale = (fingerColliderScale.val / 100f) * new Vector3(radius, 0.85f * collider.height / 2, radius);
                            n.transform.parent = child;
                            n.transform.localPosition = Vector3.zero;
                            n.transform.localRotation = Quaternion.AngleAxis(90f, Vector3.forward);
                            n.active = visualizeColliders.val;
                            colliderMarkers[collider] = n;
                        }
                        else
                            colliderMarkers[collider].transform.localScale = (fingerColliderScale.val / 100f) * new Vector3(radius, 0.85f * collider.height / 2, radius);
                    }
                }
            }
        }

        void ApplyPoseMorph()
        {
            foreach (var bone in fingerToeBones)
            {
                bool leftSideBone = bone.name.StartsWith("l");
                if ((leftSideBone && !leftSide.val) || (!leftSideBone && !rightSide.val))
                    continue;

                float bSideFactor = bone.name.StartsWith("l") ? -1 : 1;

                Quaternion targetRotation = Quaternion.identity;
                Vector3 eulerAngles = Mathf.Rad2Deg * Quaternion2Angles.GetAngles(targetRotation, bone.rotationOrder);

                float rotX = eulerAngles.x;
                float rotY = eulerAngles.y;
                float rotZ = eulerAngles.z;

                float flexOffset = bSideFactor * -fingerFlexOffset.val;

                if (bone.name.Contains("Thumb1"))
                {
                    if (fingerFlexOffset.val > 0f)
                        bone.SetBoneXRotation("IronGrip", -rotX - flexOffset / 8f);
                    else
                        bone.SetBoneXRotation("IronGrip", -rotX);
                    bone.SetBoneYRotation("IronGrip", -rotY);
                    if (fingerFlexOffset.val > 0f)
                        bone.SetBoneZRotation("IronGrip", -rotZ - flexOffset / 4f);
                    else
                        bone.SetBoneZRotation("IronGrip", -rotZ);
                }
                else if (bone.name.Contains("Thumb")) // 2 and 3 are single axis
                {
                    bone.SetBoneXRotation("IronGrip", 0f);
                    bone.SetBoneYRotation("IronGrip", -(rotY - flexOffset));
                    bone.SetBoneZRotation("IronGrip", 0f);
                }
                else if (bone.name.Contains("Toe"))
                {
                    bone.SetBoneXRotation("IronGrip", rotX);
                    bone.SetBoneYRotation("IronGrip", -rotY);
                    bone.SetBoneZRotation("IronGrip", 0f);
                }
                else // carpals and fingers
                {
                    bone.SetBoneXRotation("IronGrip", 0f);
                    if (bone.name.Contains("Carpal"))
                        bone.SetBoneYRotation("IronGrip", -rotY);
                    else if (bone.name.Contains("1"))
                    {
                        float offset = bSideFactor * fingersCloserTogether.val;
                        if (bone.name.Contains("Index") || bone.name.Contains("Pinky"))
                            offset *= 1.5f;
                        if (bone.name.Contains("Mid") || bone.name.Contains("Ring"))
                            offset = offset / 2f;
                        if (bone.name.Contains("Ring") || bone.name.Contains("Pinky"))
                            offset = -offset;

                        bone.SetBoneYRotation("IronGrip", -(rotY + offset));
                    }
                    else
                        bone.SetBoneYRotation("IronGrip", 0f); // finger joints 2 and 3 are single axis

                    bone.SetBoneZRotation("IronGrip", -(rotZ + flexOffset));
                }

                bone.SyncMorphBoneRotations(false);
            }
        }

        public void RegisterUIElements()
        {
            uiStrengthFactor = new JSONStorableFloat("Strength Factor (%)", 50f, 0.0f, 500.0f, true, true);
            uiStrengthFactor.setCallbackFunction += (float v) =>
            {
                ApplyFingerToePhysics();
                ApplyLimbPhysics();
            };
            RegisterFloat(uiStrengthFactor);

            uiLimbStrengthFactor = new JSONStorableFloat("Additional Limb Strength Multiplier (%)", 100f, 0.0f, 200.0f, true, true);
            uiLimbStrengthFactor.setCallbackFunction += (float v) =>
            {
                ApplyFingerToePhysics();
                ApplyLimbPhysics();
            };
            RegisterFloat(uiLimbStrengthFactor);

            fingersCloserTogether = new JSONStorableFloat("Fingers Closer Together", 0.0f, -50.0f, 50.0f, true, true);
            fingersCloserTogether.setCallbackFunction += (float v) =>
            {
                ApplyPoseMorph();
            };
            RegisterFloat(fingersCloserTogether);

            fingerFlexOffset = new JSONStorableFloat("Finger Flex Offset", 0.0f, -90.0f, 50.0f, true, true);
            fingerFlexOffset.setCallbackFunction += (float v) =>
            {
                ApplyPoseMorph();
            };
            RegisterFloat(fingerFlexOffset);

            fingerMassScaleFactor = new JSONStorableFloat("Finger Mass Scale (%)", 200.0f, 100.0f, 5000.0f, true, true);
            fingerMassScaleFactor.setCallbackFunction += (float scale) =>
            {
                ApplyFingerToePhysics();
            };
            RegisterFloat(fingerMassScaleFactor);

            fingerColliderScale = new JSONStorableFloat("Finger Collider Scale (%)", 100.0f, 0, 200.0f, true, true);
            fingerColliderScale.setCallbackFunction += (float scale) =>
            {
                ApplyFingerColliderSettings();
            };
            RegisterFloat(fingerColliderScale);

            fingerColliderOffset = new JSONStorableFloat("Finger Collider Offset", 0.0f, -5.0f, 5.0f, true, true);
            fingerColliderOffset.setCallbackFunction += (float Offset) =>
            {
                ApplyFingerColliderSettings();
            };
            RegisterFloat(fingerColliderOffset);

            overrideFingerPhysics = new JSONStorableBool("Control Finger/Toe Physics", true);
            overrideFingerPhysics.setCallbackFunction += (bool b) =>
            {
                if (b)
                    ApplyFingerToePhysics();
                else
                    ResetFingerToePhysics(true);
            };
            RegisterBool(overrideFingerPhysics);

            visualizeColliders = new JSONStorableBool("Visualize Colliders", false);
            visualizeColliders.setCallbackFunction += (bool b) =>
            {
                foreach (var pair in colliderMarkers)
                    pair.Value.active = b;
            };
            RegisterBool(visualizeColliders);

            useFingerDefaultsAsMin = new JSONStorableBool("Use Finger Defaults As Min", true);
            useFingerDefaultsAsMin.setCallbackFunction += (bool b) =>
            {
                ApplyFingerToePhysics();
            };
            RegisterBool(useFingerDefaultsAsMin);


            overrideHandPhysics = new JSONStorableBool("Control Wrist Physics", true);
            overrideHandPhysics.setCallbackFunction += (bool b) =>
            {
                if (b)
                    ApplyLimbPhysics();
                else
                    ResetLimbPhysics(true);
            };
            RegisterBool(overrideHandPhysics);

            overrideElbowPhysics = new JSONStorableBool("Control Elbow Physics", true);
            overrideElbowPhysics.setCallbackFunction += (bool b) =>
            {
                if (b)
                    ApplyLimbPhysics();
                else
                    ResetLimbPhysics(true);
            };
            RegisterBool(overrideElbowPhysics);

            overrideShoulderPhysics = new JSONStorableBool("Control Shoulder Physics", true);
            overrideShoulderPhysics.setCallbackFunction += (bool b) =>
            {
                if (b)
                    ApplyLimbPhysics();
                else
                    ResetLimbPhysics(true);
            };
            RegisterBool(overrideShoulderPhysics);

            leftSide = new JSONStorableBool("Affect Left Side", true);
            leftSide.setCallbackFunction += (bool b) =>
            {
                if (b)
                {
                    ApplyLimbPhysics();
                    ApplyFingerToePhysics();
                    ApplyFingerColliderSettings();
                    ApplyPoseMorph();
                }
                else
                {
                    ResetLimbPhysics(false, true);
                    ResetFingerToePhysics();
                    ResetPoseMorph();
                    ResetFingerColliderSettings();
                }
            };
            RegisterBool(leftSide);

            rightSide = new JSONStorableBool("Affect Right Side", true);
            rightSide.setCallbackFunction += (bool b) =>
            {
                if (b)
                {
                    ApplyLimbPhysics();
                    ApplyFingerToePhysics();
                    ApplyFingerColliderSettings();
                    ApplyPoseMorph();
                }
                else
                {
                    ResetLimbPhysics(false, true);
                    ResetFingerToePhysics();
                    ResetPoseMorph();
                    ResetFingerColliderSettings();
                }
            };
            RegisterBool(rightSide);

            minFingerSpring = new JSONStorableFloat("Min Finger Spring Power", 4.0f, 0.0f, 200.0f, true, true);
            minFingerSpring.setCallbackFunction += (float v) =>
            {
                fingerSpringRange.min = v;
                ApplyFingerToePhysics();
            };
            RegisterFloat(minFingerSpring);
            /*
             * 
             * Note
             * Callback dies when range is used
             * 
             * 
             */
            maxFingerSpring = new JSONStorableFloat("Max Finger Spring Power", 1000.0f, 0.0f, 10000.0f, true, true);
            maxFingerSpring.setCallbackFunction += (float v) =>
            {
                fingerSpringRange.max = v;
                ApplyFingerToePhysics();
            };
            RegisterFloat(maxFingerSpring);

            minFingerDamper = new JSONStorableFloat("Min Finger Damper", 0.1f, 0.0f, 10.0f, true, true);
            minFingerDamper.setCallbackFunction += (float v) =>
            {
                fingerDamperRange.min = v;
                ApplyFingerToePhysics();
            };
            RegisterFloat(minFingerDamper);

            maxFingerDamper = new JSONStorableFloat("Max Finger Damper", 5.0f, 0.0f, 100.0f, true, true);
            maxFingerDamper.setCallbackFunction += (float v) =>
            {
                fingerDamperRange.max = v;
                ApplyFingerToePhysics();
            };
            RegisterFloat(maxFingerDamper);

            minFingerLimitSpring = new JSONStorableFloat("Min Finger Limit Spring Power", 0f, 0.0f, 100.0f, true, true);
            minFingerLimitSpring.setCallbackFunction += (float v) =>
            {
                fingerLimitSpringRange.min = v;
                ApplyFingerToePhysics();
            };
            RegisterFloat(minFingerLimitSpring);

            maxFingerLimitSpring = new JSONStorableFloat("Max Finger Limit Spring Power", 0f, 0.0f, 10000.0f, true, true);
            maxFingerLimitSpring.setCallbackFunction += (float v) =>
            {
                fingerLimitSpringRange.max = v;
                ApplyFingerToePhysics();
            };
            RegisterFloat(maxFingerLimitSpring);

            minFingerLimitDamper = new JSONStorableFloat("Min Finger Limit Damper", 0f, 0.0f, 10.0f, true, true);
            minFingerLimitDamper.setCallbackFunction += (float v) =>
            {
                fingerLimitDamperRange.min = v;
                ApplyFingerToePhysics();
            };
            RegisterFloat(minFingerLimitDamper);

            maxFingerLimitDamper = new JSONStorableFloat("Max Finger Limit Damper", 0f, 0.0f, 100.0f, true, true);
            maxFingerLimitDamper.setCallbackFunction += (float v) =>
            {
                fingerLimitDamperRange.max = v;
                ApplyFingerToePhysics();
            };
            RegisterFloat(maxFingerLimitDamper);

            minFingerMaxForce = new JSONStorableFloat("Min Finger Maximum Force", 5f, 0.0f, 100.0f, true, true);
            minFingerMaxForce.setCallbackFunction += (float v) =>
            {
                fingerMaxForceRange.min = v;
                ApplyFingerToePhysics();
            };
            RegisterFloat(minFingerMaxForce);

            maxFingerMaxForce = new JSONStorableFloat("Max Finger Maximum Force", 100.0f, 0.0f, 10000.0f, true, true);
            maxFingerMaxForce.setCallbackFunction += (float v) =>
            {
                fingerMaxForceRange.max = v;
                ApplyFingerToePhysics();
            };
            RegisterFloat(maxFingerMaxForce);

            minHoldPositionSpring = new JSONStorableFloat("[Limbs] Min Joint Hold Position Spring", 2000f, 0.0f, 10000.0f, true, true);
            minHoldPositionSpring.setCallbackFunction += (float v) =>
            {
                holdPositionSpringRange.min = v;
                ApplyLimbPhysics();
            };
            RegisterFloat(minHoldPositionSpring);

            maxHoldPositionSpring = new JSONStorableFloat("[Limbs] Max Joint Hold Position Spring", 10000f, 0.0f, 10000.0f, true, true);
            maxHoldPositionSpring.setCallbackFunction += (float v) =>
            {
                holdPositionSpringRange.max = v;
                ApplyLimbPhysics();
            };
            RegisterFloat(maxHoldPositionSpring);

            minHoldPositionDamper = new JSONStorableFloat("[Limbs] Min Joint Hold Position Damper", 35f, 0.0f, 1000.0f, true, true);
            minHoldPositionDamper.setCallbackFunction += (float v) =>
            {
                holdPositionDamperRange.min = v;
                ApplyLimbPhysics();
            };
            RegisterFloat(minHoldPositionDamper);

            maxHoldPositionDamper = new JSONStorableFloat("[Limbs] Max Joint Hold Position Damper", 100f, 0.0f, 1000.0f, true, true);
            maxHoldPositionDamper.setCallbackFunction += (float v) =>
            {
                holdPositionDamperRange.max = v;
                ApplyLimbPhysics();
            };
            RegisterFloat(maxHoldPositionDamper);

            minHoldRotationSpring = new JSONStorableFloat("[Limbs] Min Joint Hold Rotation Spring", 100f, 0.0f, 1000.0f, true, true);
            minHoldRotationSpring.setCallbackFunction += (float v) =>
            {
                holdRotationSpringRange.min = v;
                ApplyLimbPhysics();
            };
            RegisterFloat(minHoldRotationSpring);

            maxHoldRotationSpring = new JSONStorableFloat("[Limbs] Max Joint Hold Rotation Spring", 1000f, 0.0f, 1000.0f, true, true);
            maxHoldRotationSpring.setCallbackFunction += (float v) =>
            {
                holdRotationSpringRange.max = v;
                ApplyLimbPhysics();
            };
            RegisterFloat(maxHoldRotationSpring);

            minHoldRotationDamper = new JSONStorableFloat("[Limbs] Min Joint Hold Rotation Damper", 5f, 0.0f, 10000.0f, true, true);
            minHoldRotationDamper.setCallbackFunction += (float v) =>
            {
                holdRotationDamperRange.min = v;
                ApplyLimbPhysics();
            };
            RegisterFloat(minHoldRotationDamper);

            maxHoldRotationDamper = new JSONStorableFloat("[Limbs] Max Joint Hold Rotation Damper", 1000f, 0.0f, 1000.0f, true, true);
            maxHoldRotationDamper.setCallbackFunction += (float v) =>
            {
                holdRotationDamperRange.max = v;
                ApplyLimbPhysics();
            };
            RegisterFloat(maxHoldRotationDamper);

            minJointDriveSpring = new JSONStorableFloat("[Limbs] Min Joint Drive Spring", 25f, 0.0f, 1000.0f, true, true);
            minJointDriveSpring.setCallbackFunction += (float v) =>
            {
                jointDriveSpringRange.min = v;
                ApplyLimbPhysics();
            };
            RegisterFloat(minJointDriveSpring);

            maxJointDriveSpring = new JSONStorableFloat("[Limbs] Max Joint Drive Spring", 400f, 0.0f, 1000.0f, true, true);
            maxJointDriveSpring.setCallbackFunction += (float v) =>
            {
                jointDriveSpringRange.max = v;
                ApplyLimbPhysics();
            };
            RegisterFloat(maxJointDriveSpring);

            minJointDriveDamper = new JSONStorableFloat("[Limbs] Min Joint Drive Damper", 0.5f, 0.0f, 100.0f, true, true);
            minJointDriveDamper.setCallbackFunction += (float v) =>
            {
                jointDriveDamperRange.min = v;
                ApplyLimbPhysics();
            };
            RegisterFloat(minJointDriveDamper);

            maxJointDriveDamper = new JSONStorableFloat("[Limbs] Max Joint Drive Damper", 100f, 0.0f, 100.0f, true, true);
            maxJointDriveDamper.setCallbackFunction += (float v) =>
            {
                jointDriveDamperRange.max = v;
                ApplyLimbPhysics();
            };
            RegisterFloat(maxJointDriveDamper);

            minMaxForce = new JSONStorableFloat("[Limbs] Min Joint Drive Max Force", 5f, 0.0f, 1000.0f, true, true);
            minMaxForce.setCallbackFunction += (float v) =>
            {
                jointDriveMaxForceRange.min = v;
                ApplyLimbPhysics();
            };
            RegisterFloat(minMaxForce);

            maxMaxForce = new JSONStorableFloat("[Limbs] Max Joint Drive Max Force", 400f, 0.0f, 1000.0f, true, true);
            maxMaxForce.setCallbackFunction += (float v) =>
            {
                jointDriveMaxForceRange.max = v;
                ApplyLimbPhysics();
            };
            RegisterFloat(maxMaxForce);
        }

        public void CreateUI()
        {
            var s = CreateSlider(uiStrengthFactor, false);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(uiLimbStrengthFactor, false);
            s.rangeAdjustEnabled = true;

            CreateSlider(fingerMassScaleFactor, false);
            CreateSlider(fingersCloserTogether, false);
            CreateSlider(fingerFlexOffset, false);

            CreateToggle(useFingerDefaultsAsMin, false);
            s = CreateSlider(minFingerSpring, false);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(minFingerDamper, false);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(minFingerLimitSpring, false);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(minFingerLimitDamper, false);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(minFingerMaxForce, false);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(minHoldPositionSpring, false);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(minHoldPositionDamper, false);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(minHoldRotationSpring, false);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(minHoldRotationDamper, false);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(minJointDriveSpring, false);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(minJointDriveDamper, false);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(minMaxForce, false);
            s.rangeAdjustEnabled = true;

            CreateSlider(fingerColliderScale, true);
            CreateSlider(fingerColliderOffset, true);
            CreateToggle(visualizeColliders, true);

            CreateToggle(leftSide, true);
            CreateToggle(rightSide, true);
            CreateToggle(overrideFingerPhysics, true);
            CreateToggle(overrideHandPhysics, true);
            CreateToggle(overrideElbowPhysics, true);
            CreateToggle(overrideShoulderPhysics, true);

            s = CreateSlider(maxFingerSpring, true);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(maxFingerDamper, true);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(maxFingerLimitSpring, true);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(maxFingerLimitDamper, true);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(maxFingerMaxForce, true);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(maxHoldPositionSpring, true);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(maxHoldPositionDamper, true);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(maxHoldRotationSpring, true);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(maxHoldRotationDamper, true);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(maxJointDriveSpring, true);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(maxJointDriveDamper, true);
            s.rangeAdjustEnabled = true;
            s = CreateSlider(maxMaxForce, true);
            s.rangeAdjustEnabled = true;
        }
    }
}